<?php

namespace App\Console\Commands\Bundle;

use App\Classes\Bundle;
use App\Console\Commands\Bundle\Traits\ReadsBundles;
use App\Console\Commands\Bundle\Traits\SelectsDisks;
use App\Exceptions\BundleUpdaterCannotBeFound;
use App\Exceptions\UnableToFindWikidataEntityId;
use App\Services\BundleUpdaters\Facades\BundleUpdater;
use Illuminate\Console\Command;

use function Laravel\Prompts\progress;

class Update extends Command
{
    use ReadsBundles;
    use SelectsDisks;

    /**
     * The console command description.
     *
     * @var string
     */
    protected $description = 'Update bundles extended metadata';

    protected $notUpdatable = [];

    protected $updatedCount = 0;

    protected $notUpdatedCount = 0;

    public function __construct()
    {
        $this->signature = 'bundle:update 
            { --r|recursive     : Also update sub-bundles } 
            { --source-disk=    : Use specified content disk - Defaults to <info>' . env('CONTENT_DISK') . '</info> }
            { path?             : Path to a specific bundle to update - Default to <info>/</info> }
        ';

        parent::__construct();
    }

    /**
     * Execute the console command.
     */
    public function handle()
    {
        $this->selectDisk()
            ->selectBundles()
            ->update()
            ->showReport();
    }

    /**
     * Updates specific bundle
     */
    private function handleBundle(Bundle $bundle, $progress)
    {
        $progress->hint(sprintf('Updated %s', $bundle->getPath()));

        try {
            $updater = BundleUpdater::getBundleUpdaterFor($bundle);
        } catch (BundleUpdaterCannotBeFound $ex) {
            return;
        }

        try {
            while (!$updater->canUpdateBundle()) {
                $specs = $updater->formSpecs();

                foreach ($specs as $key => $func) {
                    $result = $func();
                    $bundle->metadata()->set($key, $result);
                    $bundle->save();
                }
            }
        } catch (UnableToFindWikidataEntityId $ex) {
            $this->notUpdatable[$bundle->getPath()] = 'Missing Wikidata Entity Id';

            return;
        }

        $result = $updater->update();

        if ($result) {
            $this->updatedCount++;
        } else {
            $this->notUpdatedCount++;
        }
    }

    /**
     * Performs update on selected bundles
     */
    private function update(): self
    {
        progress('Updating bundles...', $this->bundles, function (Bundle $bundle, $progress) {
            $this->handleBundle($bundle, $progress);
        });

        return $this;
    }

    /**
     * Show a report of update procedure
     */
    private function showReport()
    {
        $this->line(sprintf('Updated bundles: <info>%d</info>', $this->updatedCount));
        $this->line(sprintf('Update not needed: <comment>%d</comment>', $this->notUpdatedCount));

        $this->reportNonUpdatable();
    }

    /**
     * Show report of non updatable bundles
     */
    private function reportNonUpdatable()
    {
        $count = count($this->notUpdatable);

        if ($count === 0) {
            return;
        }

        $this->newLine();
        $this->line(sprintf('Not updatable bundles: <comment>%d</comment>', $count));

        foreach ($this->notUpdatable as $bundle => $reason) {
            $this->newLine();
            $this->comment($bundle);
            $this->line(sprintf(' > %s', $reason));
        }
    }
}
