<?php

namespace App\Services\BundleCreators\Creators\CollectibleCreators;

use App\Classes\AttachmentsManager;
use App\Classes\Bundle;
use App\Exceptions\BundleAlreadyExists;
use App\Services\BundleCreators\Creators\BaseBundleCreator;
use App\Services\Rebrickable\RebrickableClient;
use Illuminate\Filesystem\FilesystemAdapter;
use Illuminate\Support\Str;

use function Laravel\Prompts\select;
use function Laravel\Prompts\text;

class LegoBundleCreator extends BaseBundleCreator
{
    private static string $section = 'collections';

    private static string $brand = 'lego';

    public function __construct(protected ?array $data, protected FilesystemAdapter $disk)
    {
    }

    /**
     * Create a bundle
     */
    public function createBundle(): string
    {
        $rebrickable = app()->make(RebrickableClient::class);
        $setData     = $rebrickable->getSet($this->data['rebrickableId']);
        $theme       = $rebrickable->getTheme($setData['theme_id']);
        $minifigs    = $rebrickable->getMinifigs($this->data['rebrickableId']);

        $themeSlug = Str::slug($theme['name']);
        $path      = sprintf(
            '%s/%s/%s/%s',
            static::$section,
            static::$brand,
            $themeSlug,
            $this->data['productId']
        );

        $bundle = new Bundle($path, $this->disk);

        if ($bundle->exists()) {
            throw new BundleAlreadyExists(
                sprintf('A bundle already exists in %s', $path)
            );
        }

        $bundle->markdown()->set('');

        $bundle->metadata()->setMany([
            'title'         => $setData['name'],
            'date'          => now()->toIso8601String(),
            'productId'     => $this->data['productId'],
            'rebrickableId' => $this->data['rebrickableId'],
        ]);

        $bundle->metadata('metadata')->setMany([
            'Détails' => [
                'Identifiant'      => $this->data['productId'],
                'Date de sortie'   => $setData['year'],
                'Nombre de pièces' => $setData['num_parts'],
            ],
            'Thème' => [
                $theme['name'],
            ],
        ]);

        $bundle->metadata('rebrickable')->setMany([
            'set'      => $setData,
            'minifigs' => $minifigs,
        ]);

        if (!empty($setData['set_img_url'])) {
            $ref = $bundle->attachments(AttachmentsManager::Images)->add([
                'url'         => $setData['set_img_url'],
                'attribution' => '&copy; [Rebrickable](https://rebrickable.com/)',
            ]);

            $bundle->metadata()->set('cover', $ref);
        }

        if (!empty($minifigs)) {
            foreach ($minifigs as $minifig) {
                if (empty($minifig['set_img_url'])) {
                    continue;
                }

                $bundle->attachments(AttachmentsManager::Images)->add([
                    'url'         => $minifig['set_img_url'],
                    'attribution' => '&copy; [Rebrickable](https://rebrickable.com/)',
                ]);
            }
        }

        $bundle->save();

        return $path;
    }

    /**
     * Return a boolean value indicating if the creator can actually make the
     * bundle using known data.
     */
    public function canCreateBundle(): bool
    {
        return !empty($this->data['productId']) && !empty($this->data['rebrickableId']);
    }

    /**
     * Return an array describing what kind of data the creator needs in
     * addition to the one it already has
     */
    public function formSpecs(): ?array
    {
        $specs = [];

        if (empty($this->data['productId'])) {
            $specs['productId'] = fn () => text('Product ID', '', '', true);
        } else {
            if (empty($this->data['rebrickableId'])) {
                $specs['rebrickableId'] = fn () => select(
                    'Rebrickable Set ID',
                    app()->make(RebrickableClient::class)->searchSet($this->data['productId'])
                );
            }
        }

        return $specs;
    }

    /**
     * Return a boolean value indicating if this creator in particular can
     * create bundles for specified section
     */
    public static function handles(string $section, ?array $data = []): bool
    {
        return $section === static::$section && !empty($data['brand']) && $data['brand'] === static::$brand;
    }
}
